/**
 * Copyright 2017 VMware, Inc.
 * <p>
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p>
 * https://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.micrometer.core.instrument.binder.logging;

import io.micrometer.core.Issue;
import io.micrometer.core.instrument.MeterRegistry;
import io.micrometer.core.instrument.MockClock;
import io.micrometer.core.instrument.simple.SimpleConfig;
import io.micrometer.core.instrument.simple.SimpleMeterRegistry;

import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.core.LoggerContext;
import org.apache.logging.log4j.core.config.Configuration;
import org.apache.logging.log4j.core.config.ConfigurationSource;
import org.apache.logging.log4j.core.config.Configurator;
import org.apache.logging.log4j.core.config.LoggerConfig;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.time.Duration;

import static java.util.Collections.emptyList;
import static org.assertj.core.api.Assertions.assertThat;
import static org.awaitility.Awaitility.await;

/**
 * Tests for {@link Log4j2Metrics}.
 *
 * @author Steven Sheehy
 * @author Johnny Lim
 */
class Log4j2MetricsTest {

    private final MeterRegistry registry = new SimpleMeterRegistry(SimpleConfig.DEFAULT, new MockClock());

    @BeforeEach
    void cleanUp() {
        LogManager.shutdown();
    }

    @AfterAll
    static void cleanUpAfterAll() {
        LogManager.shutdown();
    }

    @Test
    void log4j2LevelMetrics() {
        new Log4j2Metrics().bindTo(registry);

        assertThat(registry.get("log4j2.events").counter().count()).isEqualTo(0.0);

        Logger logger = LogManager.getLogger(Log4j2MetricsTest.class);
        Configurator.setLevel(Log4j2MetricsTest.class.getName(), Level.INFO);
        logger.info("info");
        logger.warn("warn");
        logger.fatal("fatal");
        logger.error("error");
        logger.debug("debug"); // shouldn't record a metric as per log level config
        logger.trace("trace"); // shouldn't record a metric as per log level config

        assertThat(registry.get("log4j2.events").tags("level", "info").counter().count()).isEqualTo(1.0);
        assertThat(registry.get("log4j2.events").tags("level", "warn").counter().count()).isEqualTo(1.0);
        assertThat(registry.get("log4j2.events").tags("level", "fatal").counter().count()).isEqualTo(1.0);
        assertThat(registry.get("log4j2.events").tags("level", "error").counter().count()).isEqualTo(1.0);
        assertThat(registry.get("log4j2.events").tags("level", "debug").counter().count()).isEqualTo(0.0);
        assertThat(registry.get("log4j2.events").tags("level", "trace").counter().count()).isEqualTo(0.0);
    }

    @Test
    void filterWhenLoggerAdditivityIsFalseShouldWork() {
        Logger additivityDisabledLogger = LogManager.getLogger("additivityDisabledLogger");
        Configurator.setLevel("additivityDisabledLogger", Level.INFO);

        LoggerContext loggerContext = (LoggerContext) LogManager.getContext();
        Configuration configuration = loggerContext.getConfiguration();
        LoggerConfig loggerConfig = configuration.getLoggerConfig("additivityDisabledLogger");
        loggerConfig.setAdditive(false);

        new Log4j2Metrics().bindTo(registry);

        assertThat(registry.get("log4j2.events").tags("level", "info").counter().count()).isEqualTo(0);

        additivityDisabledLogger.info("Hello, world!");
        assertThat(registry.get("log4j2.events").tags("level", "info").counter().count()).isEqualTo(1);
    }

    @Issue("#1466")
    @Test
    void filterWhenRootLoggerAdditivityIsFalseShouldWork() throws IOException {
        ConfigurationSource source = new ConfigurationSource(getClass().getResourceAsStream("/binder/logging/log4j2-root-logger-additivity-false.xml"));
        Configurator.initialize(null, source);

        Logger logger = LogManager.getLogger(Log4j2MetricsTest.class);

        new Log4j2Metrics().bindTo(registry);

        assertThat(registry.get("log4j2.events").tags("level", "info").counter().count()).isEqualTo(0);

        logger.info("Hello, world!");
        assertThat(registry.get("log4j2.events").tags("level", "info").counter().count()).isEqualTo(1);
    }

    @Test
    void isLevelEnabledDoesntContributeToCounts() {
        new Log4j2Metrics().bindTo(registry);

        Logger logger = LogManager.getLogger(Log4j2MetricsTest.class);
        logger.isErrorEnabled();

        assertThat(registry.get("log4j2.events").tags("level", "error").counter().count()).isEqualTo(0.0);
    }

    @Test
    void removeFilterFromLoggerContextOnClose() {
        new Log4j2Metrics().bindTo(registry);

        LoggerContext loggerContext = new LoggerContext("test");
        Log4j2Metrics log4j2Metrics = new Log4j2Metrics(emptyList(), loggerContext);
        log4j2Metrics.bindTo(registry);

        LoggerConfig loggerConfig = loggerContext.getConfiguration().getLoggerConfig(LogManager.ROOT_LOGGER_NAME);
        assertThat(loggerConfig.getFilter()).isNotNull();
        log4j2Metrics.close();
        assertThat(loggerConfig.getFilter()).isNull();
    }

    @Test
    void noDuplicateLoggingCountWhenMultipleNonAdditiveLoggersShareConfig() {
        LoggerContext loggerContext = new LoggerContext("test");

        LoggerConfig loggerConfig = new LoggerConfig("com.test", Level.INFO, false);
        Configuration configuration = loggerContext.getConfiguration();
        configuration.addLogger("com.test", loggerConfig);
        loggerContext.setConfiguration(configuration);
        loggerContext.updateLoggers();

        Logger logger1 = loggerContext.getLogger("com.test.log1");
        loggerContext.getLogger("com.test.log2");

        new Log4j2Metrics(emptyList(), loggerContext).bindTo(registry);

        assertThat(registry.get("log4j2.events").tags("level", "info").counter().count()).isEqualTo(0);
        logger1.info("Hello, world!");
        assertThat(registry.get("log4j2.events").tags("level", "info").counter().count()).isEqualTo(1);
    }

    @Issue("#2176")
    @Test
    void asyncLogShouldNotBeDuplicated() throws IOException {
        ConfigurationSource source = new ConfigurationSource(getClass().getResourceAsStream("/binder/logging/log4j2-async-logger.xml"));
        Configurator.initialize(null, source);

        Logger logger = LogManager.getLogger(Log4j2MetricsTest.class);

        new Log4j2Metrics().bindTo(registry);

        assertThat(registry.get("log4j2.events").tags("level", "info").counter().count()).isEqualTo(0);
        logger.info("Hello, world!");
        await().atMost(Duration.ofSeconds(1))
                .until(() -> registry.get("log4j2.events").tags("level", "info").counter().count() == 1);
    }

}
